#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

# Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS

# This software is published by the license of CPU-OS License, you can use and 
# distibuted this software under this License. See CPU-OS Liense for more detail.

# You should have receive a copy of CPU-OS License. If not, please contact us 
# by email <support_os@cpu-os.ac.cn>  


import shutil
from socket import timeout
import dbus
import dbus.service
import sys

import os, stat
import json
from installhandler import InstallHandler, Status
import logging
import platform
import traceback
import _thread
import urllib.request, urllib.parse, urllib.error
import configparser
import distro
import subprocess
from subprocess import Popen, PIPE

import time
import threading
from datetime import datetime
from threading import Timer

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import GLib, Gtk

from MetaRelease import MetaReleaseCore, get_current_version_with_time

# set dbus msg
global dbus_name
global dbus_path
global dbus_iface
dbus_name = "org.cdos.update"
dbus_path = "/org/cdos/update"
dbus_iface =  "org.cdos.update"

patchmgr_app_path = "/usr/share/cdos-update"
patchmgr_patches_dir = patchmgr_app_path + "/updates"
kernel_live_patches_dir = patchmgr_app_path + "/kernel_live_patches"
patches_tmp_dir = "/tmp/cdosupdate"

PATCHFILE_EXT_NAME = "cdospatch"

status_file_path = "/usr/share/cdos-update/daemon/status"
time_file_path = "/usr/share/cdos-update/daemon/time"

logging.basicConfig(format= '%(levelname)s %(asctime)-15s %(message)s file:%(filename)s line:%(lineno)d', filename='/var/log/cdosupdate-daemon.log',level=logging.DEBUG)

from software_center.server import SoftwareCenterServer
import apt_get_bash
from apt_get_bash import AptAction

import apt_pkg
from PkgCore.MyCache import MyCache
from PkgCore.UpdateList import UpdateList
from PkgCore import utils

class PatchStatus:
    INITED = 0
    DOWNLOADING = 10
    DOWNLOADED = 14
    ALLDOWNLOADED = 15
    UPDATING = 20
    VERIFYING = 22
    INSTALLING = 23
    INSTALLED = 25
    ALLINSTALLED = 30
    ERROR_INSTALL = 51

class Result:
    __error = None
    __value = None

    def __init__(self, error, value):
        self.__error = error
        self.__value = value

    def getError(self):
        return self.__error

    def getValue(self):
        return self.__value

# def checkUpdateStatus():
    # currentTime = datetime.now().strftime("%Y%m%d%H%M%S")
    # status, output = commands.getstatusoutput("check-update-status")
    # if (0 != status):
    #     logging.debug(datetime.now().strftime("%Y/%m/%d %H:%M:%S: ") + "check update status: Error")
    # else:
    #     logging.debug(datetime.now().strftime("%Y/%m/%d %H:%M:%S: ") + "check update status: Success")
    # t = Timer(7200, checkUpdateStatus)
    # t.start()

    # 如果失败，10分钟后重新执行
    # while(0 != status):
    #     logging.debug("checkUpdateStatus: reoperating")
    #     time.sleep(600)
    #     status, output = commands.getstatusoutput("check-update-status")

#check-update-status用于数据收集
def SendData():
    status, output = subprocess.getstatusoutput("check-update-status")
    if (0 != status):
        print(("Error: " + output))
    else:
        print("Success")
    statusFile = open(status_file_path, 'w')
    statusFile.write(str(status))
    statusFile.close()

#数据收集
def CheckUpdateStatusPerHour():
    print("CheckUpdateStatusPerHour")

    # 1小时之后继续检查
    t = Timer(3600, CheckUpdateStatusPerHour)

    currentDate = datetime.now().strftime("%Y%m%d")

    # 首次执行
    if (not os.path.isfile(time_file_path)):
        # print("文件不存在，创建并写入日期，发送数据")
        logging.debug(datetime.now().strftime("%Y/%m/%d %H:%M:%S") + ": 文件不存在，创建并写入日期，发送数据")
        timeFile = open(time_file_path, "w")
        timeFile.write(currentDate)
        timeFile.close()

        SendData()
        t.start()
        return

    # 判断上次是否执行成功
    statusFile = open(status_file_path, "r")
    statusOfFile = statusFile.read()
    statusOfFile = statusOfFile.strip('\n')
    statusFile.close()
    if (statusOfFile and "0" != statusOfFile):
        # print("上次执行失败，重新执行")
        logging.debug(datetime.now().strftime("%Y/%m/%d %H:%M:%S") + ": 上次执行失败，重新执行")
        SendData()
        t.start()
        return

    # 判断是否是同一天
    timeFile = open(time_file_path, "r")
    dateOfFile = timeFile.read()
    dateOfFile = dateOfFile.strip('\n')
    timeFile.close()
    # 日期不一致，写入新日期并发送数据
    if (currentDate != dateOfFile):
        # print("日期不一致，发送数据" + datetime.now().strftime("%Y%m%d%H%M"))
        logging.debug(datetime.now().strftime("%Y/%m/%d %H:%M:%S") + ": 日期不一致，写入新日期并发送数据")
        SendData()
        timeFile = open(time_file_path, "w")
        timeFile.write(currentDate)
        timeFile.close()

        t.start()

        return

    # print("日期一致，不发送" + datetime.now().strftime("%Y%m%d%H%M"))
    logging.debug(datetime.now().strftime("%Y/%m/%d %H:%M:%S") + ": 日期一致，不发送")
    t.start()

class CDOSUpdateDaemon(dbus.service.Object):

    def __init__(self):
        self.session_bus = dbus.SystemBus()
        name = dbus.service.BusName(dbus_name, bus=self.session_bus)
        dbus.service.Object.__init__(self, name, dbus_path)
        self._status = Status.INITED
        self._installedPatches = []

        self._installHandler = InstallHandler()
        self.checkInstalled()
        self.installKernelPatches()
        self.aptAction = AptAction( self.on_apt_status_changed, self.on_cache_status_changed )
        #self.soft_server = SoftwareCenterServer()#软件中心服务，可以删除不看

        self.cache = None
        self._update_query = []
        self._is_busy = False

        self.task_queue = []
        self.cur_task = None
        self.MetaReleaseCore_Obj = None
        self.old_env = {}

        self.working = False
        self.app_info = {}
        self.work_queue_pkg_name = []
        self.work_queue = []
        self.current_worker = None
        self._work_flag = False
        self._commit_flag = False
        self._work_lock = threading.Lock()

        self.backsys_name = ""
        self.patch_path_dict = {}
        self.patch_path_dict["sys_patch"] =  "/usr/share/cdos-update/nfspatch/"
        self.patch_path_dict["secure_patch"] =  "/usr/share/cdos-update/secure-nfspatch/"
        self.patch_path_dict["silence_patch"] =  "/usr/share/cdos-update/silence-nfspatch/"
        self.patch_path_dict["backup"] = "/opt/backup-recover/backup_data/"
        self.patch_path_dict["backup_dest"] = "/home/.sysupdate/"
        self.upgade_dbus_iface_obj = None
        
        self.patch_version = "/etc/os-patchversion"
        self.patch_info = "/etc/os-patchinfo"
        # self.backup_flag = False
        apt_pkg.init_config()
        #设置"Dir::Etc::sourcelist"为"/etc/apt/sources.list.d/official-package-repositories.list"
        #apt_pkg.config.set("Dir::Etc::sourcelist", "sources.list.d/official-package-repositories.list")
        apt_pkg.config.set("Dir::Etc::sourceparts", "")

        # CheckUpdateStatusPerHour()

    # system updates,获取所有更新包含系统升级，系统更新包含包更新和patch更新.
    @dbus.service.method(dbus_iface, in_signature='s', out_signature='a{sa{sv}}')
    def getAllUpdates(self, getpatchlistinfo_site_patch):
        # if working return 
        if self.working == True and (not self.app_info == {}):
            return self.app_info
        app_info = {}
        patches, secure_patches, silence_patches = self.get_latest_patch(getpatchlistinfo_site_patch)
        app_info["_patches"] = patches
        app_info["_secure_patches"] = secure_patches
        app_info["_silence_patches"] = silence_patches

        self.app_info = app_info
        return app_info

    @dbus.service.method(dbus_iface, in_signature='', out_signature='b')
    def check_env(self):
        stat = os.statvfs("/")
        used = (stat.f_blocks - stat.f_bfree) *100 / (stat.f_blocks - stat.f_bfree + stat.f_bavail) + 1
        logging.info("the root partition used %d" % used)
        if used >= 90:
            return False
        return True

    def del_sys_backupdata(self):
        if self.backsys_name == "":
            logging.error("backsys_name is empty, del failed")
            return False
        cmd = "/usr/bin/backuprecover-bin -d  " + self.backsys_name
        logging.info("del_sys_backupdata,the cmd is %s" % cmd )
        self.backsys_name = ""
        if os.system(cmd) == 0:
            return True
        else:
            logging.error("del_sys_backupdata failed")
            return False

    @dbus.service.method(dbus_iface, in_signature='sss', out_signature='b')
    def checkUpdates(self, home_dir, upgrade_site_patch, getpatchlistinfo_site_patch):
        try:
            patches_res = self.get_patchlist_to_install_with_info(getpatchlistinfo_site_patch)
            patches = patches_res.getValue()
            if len(list(patches.keys())) > 0:
                return True
        except Exception as e:
            logging.DEBUG("checkUpdates Exception:" + str(e))
            patches = {}
        #if len(patches):
        #    return True
        status , new_version, new_release = self.getSystemUpgradeState( home_dir, upgrade_site_patch)
        if status:
            return True
        deb_res, app_can_upgrade_list, app_pkgs = self.getAllDebUpdates()
        if 'num' in deb_res and deb_res['num'] > 0:
            return True
        app_info = self.get_app_info(app_pkgs)
        if len( app_info ):
            return True
        return False

    @dbus.service.method(dbus_iface, in_signature='s', out_signature='')
    def update_choice(self, choice):
        if choice == "continue":
            GLib.idle_add(self.do_work2, self.current_worker)
        else:
            self._commit_flag = False
            self.continue_work()

    # system updates
    @dbus.service.method(dbus_iface, in_signature='s', out_signature='a{sa{sv}}')
    def getSystemAndAppUpdates(self, getpatchlistinfo_site_patch):
        try:
            patches_res = self.get_patchlist_to_install_with_info(getpatchlistinfo_site_patch)
            patches = patches_res.getValue()
        except Exception as e:
            logging.DEBUG("getSystemAndAppUpdates Exception:" + str(e))
            patches = {}
        deb_res, app_can_upgrade_list, app_pkgs = self.getAllDebUpdates()
        app_info = self.get_app_info(app_pkgs)

        app_info["_patches"] = patches
        app_info["_deb_res"] = deb_res
        return app_info
        # return {"patches": patches, "deb_res": deb_res, "app_info": app_info}

    # 获取本地已经正确安装的deb包及版本的json描述(即ii状态的deb包)，通过Result结构返回结果，{"packageName":"version"}
    def get_local_package_list(self):
        deblist = subprocess.check_output(["dpkg","-l"]).decode().split('\n')
        if len(deblist) <= 5:
            return Result("no deb list", None)
        deb_desc_json = {}
        for i in range(5,len(deblist) - 1):
            deb_desc = deblist[i].split()
            if (len(deb_desc) > 0) and (deb_desc[0] == "ii"):
                deb_desc_json[deb_desc[1]] = deb_desc[2]
        return Result(None, deb_desc_json)

    #检测本地deb包列表中是否包含补丁中描述的包，以及本地deb包列表中的包版本是否在补丁描述的包版本列表中
    def check_package(self, patchinfo, local_package_info):
        try:
            # 如果没有包需要检测
            if ( (type(patchinfo) == dict) and (len(patchinfo) == 0) ) or (patchinfo == None) or (patchinfo == "null"):
                return Result(None, True)
            if type(patchinfo) != dict:
                return Result("Package version information type error", None)
            # 开始校验包版本
            for deb_name in patchinfo:
                # 如果本地deb包中没有该包
                if deb_name not in local_package_info:
                    return Result("Package version check failed", None)
                # if type(patchinfo[deb_name]) != list:
                #     pass
                # 如果本地deb包版本不在patchinfo版本列表中
                if type(patchinfo[deb_name]) == list and len(patchinfo[deb_name]) != 0 and local_package_info[deb_name] not in patchinfo[deb_name]:
                    return Result("Package version check failed", None)
            return Result(None, True)
        except Exception as ee:
            logging.error('generic exception %s:  %s' % (ee, traceback.format_exc()))
            return Result("Package version check exception", None)

    #安装补丁服务器上的补丁，返回需要安装的补丁.
    @dbus.service.method(dbus_iface, in_signature='s', out_signature='as')
    def get_patchlist_to_install(self, getpatchlistinfo_site_patch):
        try:
            patches = []
            install_patches = self.getInstalledList()
            patches_res = self.getRemotePatchlist(getpatchlistinfo_site_patch)
            all_patches = patches_res.getValue()
            patches = list(set(all_patches.keys()).difference(set(install_patches)))
            # print patches
            return patches
        except Exception as e:
            return []

    def check_patch_download(self, patch_name):
        logging.info("Begin check_patch_download:" + patch_name)
        try:
            cmd = "dpkg -l | grep ^ii | grep " + patch_name
            subprocess.check_output(cmd, shell=True)
            return True
        except Exception as e:
            return False

    def get_latest_patch(self, getpatchlistinfo_site_patch):
        try:
            cmd = "findmnt -n -F /etc/fstab --evaluate -o source /opt/backup-recover"
            dev_bakcup = os.popen(cmd).read()
            cmd = "mount " + dev_bakcup
            os.system(cmd)
            rq = urllib.request.Request( getpatchlistinfo_site_patch)
            update_info = json.load(urllib.request.urlopen(rq, timeout=3) )
            sys_ver = distro.os_release_info()['version']+ '-' + distro.os_release_info()['minor_version']
            logging.info("sys_ver="+ sys_ver)
            my_latest_patch_info = update_info[sys_ver][platform.machine()]
        except Exception as e:
            logging.error('get_latest_patch:generic exception %s:  %s' % (e, traceback.format_exc()))
            return {}, {}, {}
        logging.info("get_latest_patch: Get patch_cfg:" + str(my_latest_patch_info))

        installed_patchID, installed_secure_patchID, installed_silence_patchID = self.get_patch_id()
        logging.info("installed_patchID=" + str(installed_patchID) + " installed_secure_patchID=" + \
            str(installed_secure_patchID) + " installed_silence_patchID=" + str(installed_silence_patchID))

        patches = {}
        secure_patches = {}
        silence_patches = {}
        #sys patch
        try:
            if (installed_patchID < int(my_latest_patch_info["patches"]["patchID"])):
                patches = my_latest_patch_info["patches"]
                if self.check_patch_download(patches["latestPatchName"]):
                    patches["status"] = "download"
                else:
                    patches["status"] = ""
        except Exception as e:
            logging.error('get_latest_patch:generic exception %s:  %s' % (e, traceback.format_exc()))
            patches = {}
        #secure patch
        try:
            if (installed_secure_patchID < int(my_latest_patch_info["secure_patches"]["patchID"])):
                secure_patches = my_latest_patch_info["secure_patches"]
                if self.check_patch_download(secure_patches["latestPatchName"]):
                    secure_patches["status"] = "download"
                else:
                    secure_patches["status"] = ""
        except Exception as e:
            logging.error('get_latest_patch:generic exception %s:  %s' % (e, traceback.format_exc()))
            secure_patches = {}

        #silence patch
        try:
            if (installed_silence_patchID < int(my_latest_patch_info["silence_patches"]["patchID"])):
                silence_patches = my_latest_patch_info["silence_patches"]
                if self.check_patch_download(silence_patches["latestPatchName"]):
                    silence_patches["status"] = "download"
                else:
                    silence_patches["status"] = ""
        except Exception as e:
            logging.error('get_latest_patch:generic exception %s:  %s' % (e, traceback.format_exc()))
            silence_patches = {}

        return patches, secure_patches, silence_patches

    #向远端服务器请求获取patch的info，目前不能用，具体返回什么需要确定。
    def get_patchlist_to_install_with_info(self, getpatchlistinfo_site_patch):
        try:
            patches = []
            patch_info = {}
            install_patches = self.getInstalledList()
            patches_res = self.getRemotePatchlist(getpatchlistinfo_site_patch)
            all_patches = patches_res.getValue()
            for patchinfo in list(set(all_patches.keys()).difference(set(install_patches))):
                patch_info[patchinfo] = {}
                patch_info[patchinfo]["description"] = all_patches[patchinfo]["description"]
                patch_info[patchinfo]["size"] = str(0)
            return Result(None, patch_info)
        except Exception as e:
            return Result("get_patchlist_to_install Exception:" + str(e), [])

    #上传系统版本，kernel版本和硬件平台到服务器，获取对应的patch
    def getRemotePatchlist(self, getpatchlistinfo_site_patch):
        url = getpatchlistinfo_site_patch
        data = {"systemversion":distro.linux_distribution()[1],"kernelversion":platform.uname()[2],"hardwareplatform":platform.machine()}
        try:
            request = urllib.request.Request( url, data = json.dumps(data))
            patchinfolist = json.load(urllib.request.urlopen(request, timeout=10) )
            patch_info = {}
            # 没有本地deb列表，不更新
            local_package_info_res = self.get_local_package_list()
            if local_package_info_res.getError():
                return Result(None, [])
            local_package_info = local_package_info_res.getValue()
            for patchinfo in patchinfolist:
                check_res = self.check_package(patchinfo["package"], local_package_info)
                if check_res.getError():
                    continue
                patch_info[patchinfo["id"]] = {}
                patch_info[patchinfo["id"]]["description"] = patchinfo["description"]
                patch_info[patchinfo["id"]]["size"] = str(0)
            return Result(None, patch_info)
        except Exception as e:
            logging.error('generic exception %s:  %s' % (e, traceback.format_exc()))
            return Result("Unknown exception, please check the network", [])

    #获取远程服务器上的补丁.
    @dbus.service.method(dbus_iface, in_signature='', out_signature='a{sv}')
    def getSystemPatchUpdates(self, getpatchlistinfo_site_patch):
        res = self.getRemotePatchlist(getpatchlistinfo_site_patch)
        return res.getValue()

    # @dbus.service.method(dbus_iface, in_signature='', out_signature='a{sv}')
    def getAllDebUpdates(self):
        res = {
            "num": 0,
            "list": "",
            "size": "0B",
            "error": ""
        }
        # get the lock
        try:
            apt_pkg.pkgsystem_lock()
        except Exception as e:
            # logging.DEBUG("getAllDebUpdates Exception:" + str(e))
            pass

        try:
            if self.cache is None:
                self.cache = MyCache(None)
            else:
                self.cache.open(None)
                self.cache._initDepCache()
        except Exception as e:
            res["error"] = "getAllDebUpdates Exception:" + str(e)
            return res, [], []

        self.update_list=UpdateList(self)
        try:
            #暂时不判定哪部分属于应用.
            #system_list, system_pkgs , app_can_upgrade_list, app_pkgs = self.update_list.get_app_and_system_pkgs( self.get_all_package_names(), self.cache, eventloop_callback=None)
            system_list, system_pkgs , app_can_upgrade_list, app_pkgs = self.update_list.get_app_and_system_pkgs( [], self.cache, eventloop_callback=None)
            size = 0
            for pkg in system_pkgs:
                size += getattr(pkg.candidate, "size", 0)

            res["num"] = len(system_list)
            res["size"] = utils.humanize_size(size)
            if len(system_list) == 0:
                res["list"] = ""
            else:
                res["list"] = system_list
            return res, app_can_upgrade_list, app_pkgs

        except SystemError as e:
            # header = ("Could not calculate the upgrade")
            # desc = ("An unresolvable problem occurred while "
            #          "calculating the upgrade.\n\n"
            #          "Please report this bug against the 'update-manager' "
            #          "package and include the following error "
            #          "message:\n") + str(e)
            res["error"] = "getAllDebUpdates Exception:" + str(e)
            return res, [], []
        finally:
            try:
                apt_pkg.pkgsystem_unlock()
            except Exception as e:
                # logging.DEBUG("getAllDebUpdates Exception:" + str(e))
                pass

    # system upgrade
    @dbus.service.method(dbus_iface, in_signature='ss', out_signature='bss')
    def getSystemUpgradeState(self, home_dir, upgrade_site_patch):
        cur_release = ""
        try:
            get_release_site_patch = upgrade_site_patch + "upgrade.cfg"
            status , cur_release, new_release = self.check_need_upgrade( home_dir, get_release_site_patch)
            if status == 1:
                new_release_path = upgrade_site_patch + new_release
                new_ver = new_release.split("-")[0]
                if self.get_upgrade_configfile( home_dir, new_release_path) is True:
                    return True, new_ver, new_release
            return False, "", cur_release
        except Exception as e:
            logging.error('getSystemUpgradeState exception %s:  %s' % (e, traceback.format_exc()))
            return False, "", cur_release

    # system upgrade
    @dbus.service.method(dbus_iface, in_signature='', out_signature='bs')
    def download_upgrade_tools(self):
        try:
            # self.MetaReleaseCore_Obj.get_current_os_upgrade_info( download_meta_file_flag=fresh_meta_info)
            status, upgrade_tool_path, current_os_obj = self.MetaReleaseCore_Obj.download_upgrade_tools2()
            return status, upgrade_tool_path
        except Exception as e:
            # self.MetaReleaseCore_Obj.get_current_os_upgrade_info( download_meta_file_flag=True)
            logging.error('download_upgrade_tools exception %s:  %s' % (e, traceback.format_exc()))
            return False, ""

    @dbus.service.method(dbus_iface, in_signature='', out_signature='bs')
    def get_upgrade_tools_exits(self):
        try:
            # if current os is not get
            if self.MetaReleaseCore_Obj.current_os_obj is None:
                self.MetaReleaseCore_Obj.get_current_os_upgrade_info()
            # if self.meta_info is not None:
            return self.MetaReleaseCore_Obj.get_upgrade_tool_exist()
        except Exception as e:
            logging.error('get_upgrade_tools_exits exception %s:  %s' % (e, traceback.format_exc()))
            return False, ""

    @dbus.service.method(dbus_iface, in_signature='bs', out_signature='')
    def refreshUpgradeStatus(self, download_new_meta_data=False, upgrade_site_patch=""):
        if self.MetaReleaseCore_Obj is None:
            self.MetaReleaseCore_Obj = MetaReleaseCore( upgrade_site_patch)
        self.MetaReleaseCore_Obj.get_current_os_upgrade_info( use_new_meta_data=download_new_meta_data)

    #起一个system dbus： org.cdos.distupgrade，起了就返回True.
    @dbus.service.method(dbus_iface, in_signature='', out_signature='b')
    def get_upgrade_process_status(self):
        # ToDo: return True is upgrade is running
        try:
            _upgrade_dbus_name = "org.cdos.distupgrade"
            _upgrade_dbus_path = "/org/cdos/distupgrade"
            _upgrade_dbus_iface = "org.cdos.distupgrade"
            _tmp_dbus = dbus.SystemBus()
            _tmp_remote_obj = _tmp_dbus.get_object( _upgrade_dbus_name, _upgrade_dbus_path)
            # _tmp_iface = dbus.Interface( _tmp_remote_obj, _upgrade_dbus_iface)
            self.upgade_dbus_iface_obj = dbus.Interface( _tmp_remote_obj, _upgrade_dbus_iface)
            return True
        except Exception as e:
            logging.error( "get_upgrade_process_status: " + str(e))
            return False

    #系统升级之前的准备.
    @dbus.service.method(dbus_iface, in_signature='', out_signature='b')
    def can_dist_upgrade(self):

        #下载解压了升级工具包，执行完成prepare脚本，执行升级.
        def do_exec_dist_upgrade2(self):
            try:
                upgrade_msg_info = {}
                logging.info( "do_exec_dist_upgrade2")
                status = upgrade_tool_path = current_os_obj = None
                # if is upgrading
                if self.get_upgrade_process_status() == True:
                    try:
                        #升级工具存放目录.
                        upgrade_msg_info["upgrade_tool_dir"] = self.upgade_dbus_iface_obj.get_tool_dir()
                        #status为100后前端会执行升级.
                        self.upgradeStatusChanged(100, "Upgrading", upgrade_msg_info)
                        return
                    except Exception as e:
                        pass
                # download upgrade tools
                logging.info("download upgrade tools")
                # print self.MetaReleaseCore_Obj.get_upgrade_tool_exist()

                #判定升级工具是否已下载到本地目录.
                upgrade_tool_exist_status , upgrade_tool_path = self.MetaReleaseCore_Obj.get_upgrade_tool_exist()
                if upgrade_tool_exist_status == False:
                    self.upgradeStatusChanged(10, "Fetching upgrade tools", upgrade_msg_info)
                    status, upgrade_tool_path, current_os_obj = self.MetaReleaseCore_Obj.download_upgrade_tools2()

                time.sleep(1)
                self.upgradeStatusChanged(50, "Extracting upgrade tools", upgrade_msg_info)
                upgrade_tool_dir = self.MetaReleaseCore_Obj.extract_upgrade_tools()
                upgrade_msg_info["upgrade_tool_dir"] = upgrade_tool_dir
                time.sleep(1)
                self.upgradeStatusChanged(80, "Prepare for upgrade", upgrade_msg_info)
                prepare_script = os.path.join(upgrade_msg_info["upgrade_tool_dir"], "prepare")
                #执行升级工具包中的prepare脚本，拷贝dbus和翻译相关配置.
                if os.system("sudo bash " + prepare_script) == 0:
                    self.upgradeStatusChanged(100, "Upgrading", upgrade_msg_info)
                else:
                    self.upgradeStatusChanged(101, "Prepare upgrade error", {})
                    logging.error("do_exec_dist_upgrade2: run preprare scripts error")

            except Exception as e:
                logging.error('do_exec_dist_upgrade2 exception %s:  %s' % (e, traceback.format_exc()))
                self.upgradeStatusChanged(101, "Prepare upgrade error", {})

        try:
            _thread.start_new_thread(do_exec_dist_upgrade2, (self,))
            return True
        except Exception as e:
            logging.error('can_dist_upgrade exception %s:  %s' % (e, traceback.format_exc()))
            return False

    def _backup_env(self, env):
        for key in env:
            oldval=None
            try:
                if key in os.environ:
                    oldval = os.environ[key]
            except Exception as e:
                logging.error('_backup_env exception %s:  %s' % (e, traceback.format_exc()))
                pass
            self.old_env[key]=oldval
            os.environ[key] = env[key]

    def _restore_env(self):
        for key in self.old_env:
            if self.old_env[key] is not None :
                os.environ[key] = self.old_env[key]
            else:
                del os.environ[key]

    # system upgrade
    @dbus.service.method(dbus_iface, in_signature='sb', out_signature='bss')
    def getSystemUpgradeState2(self, upgrade_site_patch, use_new_meta_data = False):
        cur_release = ""
        try:
            cur_release = get_current_version_with_time()
            # self.MetaReleaseCore_Obj = MetaReleaseCore( upgrade_site_patch)
            if self.get_upgrade_process_status() == False:
                self.refreshUpgradeStatus( download_new_meta_data=True, upgrade_site_patch=upgrade_site_patch)
            else:
                self.refreshUpgradeStatus( upgrade_site_patch=upgrade_site_patch)
            self.MetaReleaseCore_Obj.get_current_os_upgrade_info(use_new_meta_data=use_new_meta_data)
            if ( self.MetaReleaseCore_Obj.current_os_obj is not None) and ( self.MetaReleaseCore_Obj.current_os_obj.supported):
                    return True, self.MetaReleaseCore_Obj.current_os_obj.dist_version_id, self.MetaReleaseCore_Obj.current_os_obj.dist_version
            return False, "", cur_release
        except Exception as e:
            logging.error('getSystemUpgradeState2 exception %s:  %s' % (e, traceback.format_exc()))
            return False, "", cur_release

    def get_upgrade_configfile(self,  home_dir, new_release_path):
        origin_manifest = {}
        new_manifest = {}
        cur_manifest = {}
        server_path = new_release_path
        if not server_path.endswith("/"):
            server_path += "/"
        local_path =os.path.join( home_dir, ".config/cdosupgrade/")
        grub_efimode_remove_list = ["grub-pc", "grub-pc-bin", "grub-gfxpayload-lists"]
        grub_efimode_add_list = ["mokutil", "grub-efi-amd64", "efibootmgr", "shim", "shim-signed", "grub-efi-amd64-bin", "grub-efi-amd64-signed"]
        if not os.path.exists(local_path):
            os.makedirs(local_path,mode=0o755)
        try:
            urllib.request.urlretrieve(server_path+"filesystem.manifest",local_path+"filesystem.manifest")
            urllib.request.urlretrieve(server_path+"official-package-repositories.list",local_path+"official-package-repositories.list")
            urllib.request.urlretrieve(server_path+"filesystem.manifest-remove",local_path+"filesystem.manifest-remove")
            urllib.request.urlretrieve(server_path+"os-release",local_path+"os-release")
            urllib.request.urlretrieve(server_path+"lsb-release",local_path+"lsb-release")
            urllib.request.urlretrieve(server_path+"fix_bug.py",local_path+"fix_bug.py")
        except Exception as e:
            logging.error('get_upgrade_configfile exception %s:  %s' % (e, traceback.format_exc()))
            return False
        origin_manifest_path = ("/var/log/installer/filesystem.manifest")
        if os.path.exists(origin_manifest_path):
            with open(origin_manifest_path) as f:
                for line in f.readlines():
                    result = line.strip("\n").split()
                    if len(result) >= 2:
                        key = result[0]
                        value = result[1]
                        origin_manifest[key] = value
        cur_manifest_path = os.path.join( home_dir, ".config/cdosupgrade/current.manifest")
        os.system("dpkg-query -W --showformat='${Package} ${Version}\n' > %s" % cur_manifest_path)
        with open(cur_manifest_path) as f:
            for line in f.readlines():
                result = line.strip("\n").split()
                if len(result) >= 2:
                     key = result[0]
                     value = result[1]
                     cur_manifest[key] = value
        if os.path.exists("/sys/firmware/efi"):
            for package in grub_efimode_add_list:
                new_manifest[package] = "0.0.1"
        new_manifest_path = os.path.join( home_dir, ".config/cdosupgrade/filesystem.manifest")
        with open(new_manifest_path) as f:
            for line in f.readlines():
                result = line.strip("\n").split()
                if len(result) >= 2:
                     key = result[0]
                     value = result[1]
                     if os.path.exists("/sys/firmware/efi") and key in grub_efimode_remove_list:
                         continue
                     new_manifest[key] = value
        custom_pkgs = []
        nhold_pkgs = []
        upgrade_pkgs = []
        if os.path.exists(origin_manifest_path):
            custom_pkgs = set(cur_manifest.keys()) - set(origin_manifest.keys())
            nhold_pkgs = set(origin_manifest.keys()) - set(new_manifest.keys()) - custom_pkgs
        new_pkgs = set(new_manifest.keys()) - set(cur_manifest.keys())
        upgrade_pkgs = set(new_manifest.keys()) - new_pkgs
        upgrade_pkgs_version = []
        upgrade_manifest_path = os.path.join( home_dir, ".config/cdosupgrade/upgrade_pkgs")
        for package in upgrade_pkgs:
            if cur_manifest[package] == new_manifest[package]:
                continue
            else:
                item = "%s %s" %(package, new_manifest[package])
                upgrade_pkgs_version.append(item)
        with open(upgrade_manifest_path, 'w') as f:
            f.write('\n'.join(upgrade_pkgs_version))
        install_pkgs_list = []
        install_manifest_path = os.path.join( home_dir, ".config/cdosupgrade/new_pkgs")
        for package in new_pkgs:
            item = "%s %s" %(package, new_manifest[package])
            install_pkgs_list.append(item)
        with open(install_manifest_path, 'w') as f:
            f.write('\n'.join(install_pkgs_list))
        return True

    def compare_version(self, cur_release, new_release):
        try:
            cur_release_date = cur_release.split('-')[2]
            new_release_date = new_release.split('-')[2]
        except:
            return 2
        if cur_release_date.isdigit() and new_release_date.isdigit():
            if new_release_date > cur_release_date:
                return 1
            else:
                return 0
        else:
            return 2

    def check_need_upgrade(self, home_dir, get_release_site_patch):
        try:
            new_release_file = get_release_site_patch
            local_conf_dir = os.path.join(home_dir, ".config/cdosupgrade/")
            if not os.path.exists(local_conf_dir):
                os.mkdir(local_conf_dir)
            local_conf_file = os.path.join(home_dir, ".config/cdosupgrade/upgrade.cfg")
            conf = configparser.ConfigParser(defaults=None)
            try:
                urllib.request.urlretrieve(new_release_file,local_conf_file)
                conf.read(local_conf_file)
            except Exception as e:
                logging.DEBUG("check_need_upgrade Exception:" + str(e))
                return 3 ,"" ,""
            new_release = conf.get("lastest", "OS_VERSION")
            with open ("/etc/os-release") as f:
                for line in f.readlines():
                    if line.startswith("VERSION="):
                        cur_release = eval(line.split("=")[-1].strip("\n"))
            return self.compare_version(cur_release, new_release), cur_release, new_release
        except Exception as e:
            logging.DEBUG("check_need_upgrade Exception:" + str(e))
            return 3 ,"" ,""

    #通过对比"kpatch list"出来的和"/usr/share/cdos-update/kernel_live_patches"目录下的补丁目录文件，获取没有load的热补丁，通过kpatch load安装上.
    @dbus.service.method(dbus_iface, in_signature='', out_signature='')
    def installKernelPatches(self):
        logging.debug("installKernelPatches")
        # 初始化热补丁目录
        if os.path.exists(kernel_live_patches_dir) and os.path.isfile(kernel_live_patches_dir):
            os.remove(kernel_live_patches_dir)
        if not os.path.exists(kernel_live_patches_dir):
            os.makedirs(kernel_live_patches_dir, 0o644)
        deb_info = self.get_deb_info()
        if deb_info == {}:
            return
        # 获取已安装列表
        cmd = "kpatch list|grep kpatch"
        subp = Popen(cmd, shell=True, stdout=PIPE)
        rc = subp.wait()
        installed_kernel_ids_str = subp.stdout.read().strip().decode().split('\n')
        # 已安装列表
        installed_kernel_ids = []
        for installed_kernel_id_str in installed_kernel_ids_str:
            if not ( installed_kernel_id_str == ''):
                installed_kernel_ids.append( installed_kernel_id_str[7:])
        # 获取本地待安装内核热补丁
        for kernel_live_patch in os.listdir(kernel_live_patches_dir):
            # 判断内核热补丁合法性
            kernel_live_patch_path = os.path.join(kernel_live_patches_dir, kernel_live_patch)
            kernel_live_patch_path_json = kernel_live_patch_path+".json"
            if os.path.isfile(kernel_live_patch_path) and os.path.isfile(kernel_live_patch_path_json) and kernel_live_patch.startswith("kpatch-") and kernel_live_patch.endswith(".ko"):
                try:
                    # 获取内核热补丁id
                    kernel_id = kernel_live_patch[7:-3]
                    # 如果该热补丁未被安装
                    if (kernel_id not in installed_kernel_ids) and (self.check_kernel(kernel_live_patch_path_json, deb_info) == True):
                        # 安装待安装的内核热补丁
                        os.system("sudo kpatch load " + kernel_live_patch_path)
                except Exception as e:
                    logging.error('Exception %s:  %s' % (e, traceback.format_exc()))
                    continue
        return

    # 获取本地已经正确安装的deb包及版本的json描述(即ii状态的deb包),返回{"packageName": "version"}
    def get_deb_info(self):
        deblist = subprocess.check_output(["dpkg","-l"]).decode().split('\n')
        if len(deblist) <= 5:
            return {}
        deb_desc_json = {}
        for i in range(5,len(deblist) - 1):
            deb_desc = deblist[i].split()
            if (len(deb_desc) > 0) and (deb_desc[0] == "ii"):
                deb_desc_json[deb_desc[1]] = deb_desc[2]
        return deb_desc_json

    #检测内核热补丁是否适合当前的系统
    def check_kernel(self, kernel_json, deb_desc_json):
        try:
            kernel_json_file = file(kernel_json)
            kernel_info = json.load(kernel_json_file)
            os_kernel_release = os.uname()[2]
            if not os_kernel_release == kernel_info["kernel_release"]:
                return False
            # 开始校验包版本
            package_info = kernel_info["packages"]
            for deb_name in package_info:
                # 如果本地deb包中没有该包
                if deb_name not in deb_desc_json:
                    return False
                if not type(package_info[deb_name]) == list :
                    return False
                if len(package_info[deb_name]) != 0 and deb_desc_json[deb_name] not in package_info[deb_name]:
                    return False
            return True
        except Exception as e:
            return False

    @dbus.service.method(dbus_iface, in_signature='', out_signature='s')
    def getVersion(self):
        logging.debug("getVersion 1.1.3")
        return "1.1.3"

    #查看/usr/share/cdos-update/updates目录下的补丁文件获取已经安装的补丁数目
    @dbus.service.method(dbus_iface, in_signature='', out_signature='i')
    def checkInstalled(self):
        logging.debug("checkInstalled")
        try:
            alist = []
            if not os.path.exists(patchmgr_patches_dir):
                os.makedirs(patchmgr_patches_dir, 0o644)
            for afile in os.listdir(patchmgr_patches_dir):
                if os.path.isfile(os.path.join(patchmgr_patches_dir, afile)):
                    if afile.endswith("." + PATCHFILE_EXT_NAME):
                        alist.append(afile[:-(1+len(PATCHFILE_EXT_NAME))])
            self._installedPatches = alist
        except Exception as e:
            logging.error('Exception %s:  %s' % (e, traceback.format_exc()))
            return -1
        return len(self._installedPatches)

    #查看/usr/share/cdos-update/updates目录下的补丁文件获取已经安装的补丁列表
    @dbus.service.method(dbus_iface, in_signature='', out_signature='as')
    def getInstalledList(self):
        logging.debug("getInstalledList")
        try:
            alist = []
            if not os.path.exists(patchmgr_patches_dir):
                return alist
            for afile in os.listdir(patchmgr_patches_dir):
                if os.path.isfile(os.path.join(patchmgr_patches_dir, afile)):
                    if afile.endswith("." + PATCHFILE_EXT_NAME):
                        alist.append(afile[:-(1+len(PATCHFILE_EXT_NAME))])
            return alist
        except Exception as e:
            logging.error('Exception %s:  %s' % (e, traceback.format_exc()))
        return []

    @dbus.service.method(dbus_iface, in_signature='', out_signature='as')
    def getDownloadedList(self):
        logging.debug("getDownloadedList")
        downloadedList = []
        try:
            if not os.path.exists(patches_tmp_dir):
                return downloadedList
            for afile in os.listdir(patches_tmp_dir):
                if os.path.isfile(os.path.join(patches_tmp_dir, afile)):
                    if afile.endswith("." + PATCHFILE_EXT_NAME):
                        downloadedList.append(afile[:-(1+len(PATCHFILE_EXT_NAME))])
            return downloadedList
        except Exception as e:
            logging.error('Exception %s:  %s' % (e, traceback.format_exc()))
        return downloadedList

    def progressCallback(self, statusCode, status_text, patchID, correct_install_patches_list, patchList, progress, paramList={}):
        logging.info("progressCallback")
        if statusCode == Status.INSTALLED:
            self._installedPatches.append(patchID)
        if statusCode == Status.ALLINSTALLED:
            self.checkInstalled()
            if self.backsys_name != "":
                os.remove("/usr/share/cdos-update/update_flag")
                self.del_sys_backupdata()
        if statusCode == Status.ERROR_INSTALL:
            self.checkInstalled()
        self.statusChanged(statusCode, status_text, patchID, correct_install_patches_list, patchList, progress, paramList)

    @dbus.service.method(dbus_iface, in_signature='as', out_signature='ba{sa{sv}}')
    def verify_offline_patch(self, patch_name_list):
        logging.info("Begin verify patch list:" + str(patch_name_list))
        sys_patches = {"patch_name": "", "patch_id":0, "import_patch_list": [""]}
        secure_patches = {"patch_name": "", "patch_id":0, "import_patch_list": [""]}
        fail_res={"_patches" : sys_patches, "_secure_patches": secure_patches}
        sys_ver = distro.os_release_info()['minor_version'].lower()
        sys_patch_id_list = []
        secure_patch_id_list = []
        sys_patch_id_name = {}
        secure_patch_id_name = {}
        for patch_name in patch_name_list:
            #verify patchname
            if (not sys_ver in patch_name) or (not "nfsupdate" in patch_name)  or (not ".deb" in patch_name):
                logging.error("verify_offline_patch: verify failed with " + patch_name)
                return False, fail_res
            try:
                id = int(patch_name.split('_')[0].split('-')[-1])
                if "secure" in patch_name:
                    secure_patch_id_list.append(id)
                    secure_patch_id_name[id] = patch_name
                else:
                    sys_patch_id_list.append(id)
                    sys_patch_id_name[id] = patch_name
            except Exception as e:
                logging.error("verify_offline_patch: verify failed with " + patch_name + " exception:" + str(e))
                return False, fail_res
        logging.info("verify_offline_patch: sys patch:" + str(sys_patch_id_name) + "secure patch:" + str(secure_patch_id_name))

        if len(sys_patch_id_list) != len(sys_patch_id_name) or len(secure_patch_id_list) != len(secure_patch_id_name):
            logging.error("verify_offline_patch verify failed, maybe duplicate patch" )
            return False, fail_res

        sys_patch_id_list.sort()
        secure_patch_id_list.sort()

        #Verify that is installed
        latest_sys_patchID, latest_secure_patchID, latest_silence_patchID = installed_patchID, installed_secure_patchID, installed_silence_patchID = self.get_patch_id()
        if not ((len(sys_patch_id_list) != 0 and installed_patchID < sys_patch_id_list[-1]) \
            or (len(secure_patch_id_list) != 0 and installed_secure_patchID < secure_patch_id_list[-1])):
            logging.error("verify_offline_patch: verify failed, no updated patches, installed_patchID=" + str(installed_patchID) + "  installed_secure_patchID=" + str(installed_secure_patchID))
            return False, fail_res

        #校验补丁是否连续
        import_sys_patch_list = []
        import_secure_patch_list = []
        for i in sys_patch_id_list:
            if i <= latest_sys_patchID:
                continue
            elif i == latest_sys_patchID + 1:
                latest_sys_patchID = latest_sys_patchID + 1
                import_sys_patch_list.append(sys_patch_id_name[latest_sys_patchID])
            else:
                logging.error("verify_offline_patch: verify failed,illegal patch sequence:" + str(sys_patch_id_list))
                return False, fail_res
        for i in secure_patch_id_list:
            if i <= latest_secure_patchID:
                continue
            elif i == latest_secure_patchID + 1:
                latest_secure_patchID = latest_secure_patchID + 1
                import_secure_patch_list.append(secure_patch_id_name[latest_secure_patchID])
            else:
                logging.error("verify_offline_patch: verify failed,illegal patch sequence:" + str(secure_patch_id_list))
                return False, fail_res

        if len(import_sys_patch_list) != 0:
            sys_patches["patch_name"] = sys_patch_id_name[latest_sys_patchID].split('_')[0]
            sys_patches["patch_id"] = latest_sys_patchID 
            sys_patches["import_patch_list"] = import_sys_patch_list
        if len(import_secure_patch_list) != 0:
            secure_patches["patch_name"] = secure_patch_id_name[latest_secure_patchID].split('_')[0]
            secure_patches["patch_id"] = latest_secure_patchID
            secure_patches["import_patch_list"] = import_secure_patch_list
        logging.info("verify_offline_patch: sys_patches:" + str(sys_patches) + "secure_patches" + str(secure_patches))
        return True, {"_patches" : sys_patches, "_secure_patches": secure_patches}


    @dbus.service.method(dbus_iface, in_signature='a{sv}s', out_signature='')
    def download_patch(self, env, patch_name):
        logging.info("download: %s" % patch_name)
        worker = {}
        worker["env"] = env
        worker["action"] = "download_patch"
        if "secure" in patch_name:
            worker["type"] = "secure_patch"
        elif "silence" in patch_name:
            worker["type"] = "silence_patch"
        else:
            worker["type"] = "sys_patch"
        worker["install_patch_name"] = patch_name
        # add for mark
        worker["pkg_name"] = "system_patch_update"
        worker["install_pkg_list"] = patch_name
        worker["pkg_version"] = ""
        worker["callback"] = None
        self.push_work(worker)

    @dbus.service.method(dbus_iface, in_signature='a{sv}', out_signature='')
    def import_offline_patch(self, env):
        logging.info("import_offline_patch: set task")
        worker = {}
        env_daemon = os.getenv("PATH")
        if env_daemon != env["PATH"]:
            env["PATH"] = env_daemon
        worker["env"] = env
        worker["action"] = "import_offline_patch"
        worker["type"] = "system_patch_update"
        # add for mark
        worker["pkg_name"] = "system_patch_update"
        worker["pkg_version"] = ""
        worker["callback"] = None
        self.push_work(worker)

    @dbus.service.method(dbus_iface, in_signature='sas', out_signature='s')
    def download(self, getpatch_site_path, patchIDs):
        logging.info("download: %s %s" % (getpatch_site_path, str(patchIDs)) )
        if self._status == Status.INITED or self._status == Status.ALLDOWNLOADED or self._status == Status.ALLINSTALLED or self._status == Status.ERROR_INSTALL:
            patchIDList = []
            for i in range(len(patchIDs)):
                patchIDList.append(str(patchIDs[i]))
            try:
                if self._installHandler.start_Update(getpatch_site_path, patchIDList, self.progressCallback, True):
                    return "Started"
            except Exception as e:
                logging.error('Exception %s:  %s' % (e, traceback.format_exc()))
            return "Failed"
        else:
            return "Failed"

    @dbus.service.method(dbus_iface, in_signature='is', out_signature='s')
    def get_offline_patch_desc(self,patch_id, type):
        logging.info("will get offline patch desc with " + type)
        desc = ""
        try:
            dir = self.patch_path_dict[type] + "patch"
        except Exception as e:
            logging.error('Exception %s:  %s' % (e, traceback.format_exc()))
            dir = self.patch_path_dict["sys_patch"] + "patch"
        installed_sys_patchID, installed_secure_patchID, installed_silence_patchID= self.get_patch_id()
        if type == "sys_patch":
            installed_patchID = installed_sys_patchID
        elif type == "secure_patch":
            installed_patchID = installed_secure_patchID
        else:
            installed_patchID = installed_silence_patchID
        while (patch_id > installed_patchID):
            if desc != "":
                desc = desc + "\n"
            try:
                conf = dir + str(patch_id) + "/conf.json"
                with open(conf) as cfg:
                        patch_info = json.load(cfg)
                logging.info("get_offline_patch_desc: get patch info:" + str(patch_info))
                desc = desc +  patch_info['desc']
            except Exception as e:
                logging.error("get_offline_patch_desc: get patch info failed,error:" + str(e))

            finally:
                patch_id = patch_id - 1
        logging.info("get offline patch desc:" + desc)
        return desc


    @dbus.service.method(dbus_iface, in_signature='', out_signature='')
    def sys_reboot(self):
        logging.info("will reboot sys")
        os.system("sync;sync;systemctl reboot")

    @dbus.service.method(dbus_iface, in_signature='', out_signature='b')
    def recover(self):
        logging.info("recover: begin")
        #if self.backup_flag and os.path.exists(self.patch_path_dict["backup"] + "backup_sysupdate/"):
        if os.path.exists(self.patch_path_dict["backup_dest"]):
            worker = {}
            worker["action"] = "recover"
            worker["pkg_name"] = "recover"
            self.push_work(worker)
            return True
        else:
            return False
    
    def get_patch_package(self):
        try:
            cur_sysupdate_patch_info_file = self.patch_info
            cur_sysupdate_patch_name = ""
            cur_sysupdate_patch_lines = ""
            if os.path.exists(cur_sysupdate_patch_info_file):
                with open(cur_sysupdate_patch_info_file, 'r') as f:
                    cur_sysupdate_patch_lines = f.readlines() 
                    if (len(cur_sysupdate_patch_lines) == 0):
                        logging.debug("patch info is empty.")
                        return ""
                    else:
                        cur_sysupdate_patch_name = cur_sysupdate_patch_lines[-1].split("++")[0]
                        return cur_sysupdate_patch_name
            else:
                logging.debug("not exist patch info")
                return ""
        except Exception as e:
            logging.error("remove_patch_package failed " + str(e))
            return ""

    def remove_patch_package(self, patch_package_name):
        try:
            cmd = "DEBIAN_FRONTEND=noninteractive apt-get purge -y " + patch_package_name + " > /dev/null 2>&1"
            if os.system(cmd) != 0:
                logging.error(cmd + " failed!")
            else:
                logging.info(cmd + " success!")
        except Exception as e:
            logging.error("remove_patch_package failed " + str(e))
            return
    
    def do_recover(self):
        logging.info("do_recover: begin")
        patch_package = ""
        msg = { 'status': 'inprogress', 'percent': 0, 'action':'recover'}
        self.backup_recover_status_changed(msg)
        if not os.path.exists(self.patch_path_dict["backup_dest"]):
            logging.error("do_recover failed")
            msg = { 'status': 'failed', 'percent': 100, 'action':'recover'}
            self.backup_recover_status_changed(msg)
        patch_package = self.get_patch_package()
        cmd = "/usr/bin/backuprecover-bin  -r  backup_sysupdate  " + self.patch_path_dict["backup"] + "var/sysupdate_exclude.list " + "-p " + self.patch_path_dict["backup_dest"]
        logging.info("do_recover: cmd:" + cmd)
        if os.system(cmd) != 0:
            logging.error("do_recover failed, will try again ")
            if os.system(cmd) != 0:
                logging.error("do_recover failed again")
                msg = { 'status': 'failed', 'percent': 100, 'action':'recover'}
                self.backup_recover_status_changed(msg)
            else:
                logging.info("do_recover success")
                self.remove_patch_package(patch_package)
                msg = { 'status': 'success', 'percent': 100, 'action':'recover'}
                self.backup_recover_status_changed(msg)
        else:
            logging.info("do_recover success")
            self.remove_patch_package(patch_package)
            msg = { 'status': 'success', 'percent': 100, 'action':'recover'}
            self.backup_recover_status_changed(msg)
    
    @dbus.service.method(dbus_iface, in_signature='', out_signature='as')
    def check_backup_data_exists(self):
        logging.info("check__backup_data_exists: begin")
        try:
            if not os.path.exists(self.patch_path_dict["backup_dest"] + 'backup_sysupdate'):
                logging.error("there are not exist .sysupdate/backup_sysupdate ,so return")
                return []

            #备份的系统安装补丁包描述信息，用于判断与当前的信息是否一致
            back_sysupdate_patch_info_file = self.patch_path_dict["backup_dest"]  + "backup_sysupdate/" +  self.patch_info
            back_sysupdate_patch_lines = ""
            if os.path.exists(back_sysupdate_patch_info_file):
                with open(back_sysupdate_patch_info_file, 'r') as f:
                    back_sysupdate_patch_lines = f.readlines()
                    if (len(back_sysupdate_patch_lines) == 0) :
                        back_sysupdate_patch_lines = "unknow"
            else:
                back_sysupdate_patch_lines = "unknow"

            #最后一次系统安装补丁包描述信息
            cur_patch_info_record_file = self.patch_info
            cur_sysupdate_patch_name = ""
            cur_sysupdate_patch_desc = ""
            cur_sysupdate_patch_lines = ""
            if os.path.exists(cur_patch_info_record_file):
                with open(cur_patch_info_record_file, 'r') as f:
                    cur_sysupdate_patch_lines = f.readlines()
                    if (len(cur_sysupdate_patch_lines) == 0):
                        return []
                    elif (str(cur_sysupdate_patch_lines) == str(back_sysupdate_patch_lines)) :
                        return []
                    else:
                        cur_sysupdate_patch_name = cur_sysupdate_patch_lines[-1].split("++")[0]
                        cur_sysupdate_patch_desc = cur_sysupdate_patch_lines[-1].split("++")[1].strip()
            else:
                return []

            #最后一次系统更新时间
            backupRecords_path = self.patch_path_dict["backup"] + "var/backupRecords"
            if os.path.exists(backupRecords_path):
                with open(backupRecords_path, 'r') as f:
                    lines = f.readlines()
                    time_last_line = lines[-1]
                last_sysupdate_time = time_last_line.split(" ")[1]
                last_sysupdate_day = last_sysupdate_time.split("_")[0]
            else:
                last_sysupdate_day = " "
            logging.debug("patch_name:" + cur_sysupdate_patch_name + ", " + "last_sysupdate_patch_desc:" + cur_sysupdate_patch_desc + "last_sysupdate_time:" + last_sysupdate_day )
            return [cur_sysupdate_patch_name, cur_sysupdate_patch_desc, last_sysupdate_day]
        except Exception as e:
            logging.error("read backup data info fail" + str(e))
            return []
    
    @dbus.service.method(dbus_iface, in_signature='', out_signature='b')
    def backup(self):
        logging.info("backup: begin")
        try:
            if not os.access("/usr/bin/backuprecover-bin",os.X_OK):
                logging.error("there not exist /usr/bin/backuprecover-bin, so return")
                return False
            if not os.path.exists(self.patch_path_dict["backup"] + "var/sysupdate_exclude.list"):
                logging.error("there not exist sysupdate_exclude.list ,so return")
                return False
            sys_space_info = os.statvfs("/")
            sys_space_used = int((sys_space_info.f_blocks - sys_space_info.f_bfree) * sys_space_info.f_bsize / 1024 / 1024)
            home_space_info = os.statvfs("/home/")
            backup_space_avail = int(home_space_info.f_bavail * home_space_info.f_bsize / 1024 / 1024)
            try:
                if (os.path.exists(self.patch_path_dict["backup_dest"] + 'backup_sysupdate')) and (os.path.exists(self.patch_path_dict["backup"] + "var/log/backup_sysupdate_backup.log")):
                ###计算现有备份占用空间MB
                    cmd = 'grep -E "total *size *is" ' + self.patch_path_dict["backup"] + 'var/log/backup_sysupdate_backup.log | awk \'{print $4}\''
                    temp = subprocess.check_output(cmd , shell=True)[:-2]
                    backup_space_used = int(float(temp) * 1024)
                else:
                    backup_space_used = 0
            except Exception as e:
                backup_space_used = 0
                logging.error('Exception %s:  %s' % (e, traceback.format_exc()))
            logging.info("sys_use:" + str(sys_space_used) + " avail:" + str(backup_space_avail) + " old_backup_use:" + str(backup_space_used))
            if (sys_space_used >= (backup_space_avail + backup_space_used)):
                logging.error("backup: There is not enough space for backups ")
                return False
            worker = {}
            worker["action"] = "backup"
            worker["pkg_name"] = "backup"
            self.push_work(worker)  
            return True
        except Exception as e:
            logging.error('Exception %s:  %s' % (e, traceback.format_exc()))
            return False
    
    def do_backup(self):
        logging.info("do_backup: begin")
        if not (os.path.exists(self.patch_path_dict["backup_dest"])):
            os.mkdir(self.patch_path_dict["backup_dest"])
            os.chmod(self.patch_path_dict["backup_dest"], stat.S_IRWXU)
        if not (os.path.exists(self.patch_path_dict["backup"] + "var/log/")):
            os.mkdir(self.patch_path_dict["backup"] + "var/log/")
        msg = { 'status': 'inprogress', 'percent': 0, 'action':'backup'}
        self.backup_recover_status_changed(msg)
        if (os.path.exists(self.patch_path_dict["backup"] + "backup_sysupdate/")):
        	cmd = "/usr/bin/backuprecover-bin  -ib  backup_sysupdate  " + self.patch_path_dict["backup"] + "var/sysupdate_exclude.list " + "-p " + self.patch_path_dict["backup_dest"]
        else:
        	cmd = "/usr/bin/backuprecover-bin  -b  backup_sysupdate  " + self.patch_path_dict["backup"] + "var/sysupdate_exclude.list " + "-p " + self.patch_path_dict["backup_dest"]
        logging.info("do_backup: cmd:" + cmd)
        if os.system(cmd) != 0:
            logging.error("backup failed, will try again")
            if os.system(cmd) != 0:
                logging.error("backup failed again")
                msg = { 'status': 'failed', 'percent': 100, 'action':'backup'}
                self.backup_recover_status_changed(msg)
            else:
                # self.backup_flag = True
                logging.info("backup success")
                msg = { 'status': 'success', 'percent': 100, 'action':'backup'}
                self.backup_recover_status_changed(msg)
        else:
            # self.backup_flag = True
            logging.info("backup success")
            msg = { 'status': 'success', 'percent': 100, 'action':'backup'}
            self.backup_recover_status_changed(msg)
        
    
    @dbus.service.method(dbus_iface, in_signature='a{sv}si', out_signature='')
    def install_patch(self, env, patch_name, patchID):
        worker = {}
        worker["env"] = env
        worker["action"] = "install_patch"
        if "secure" in patch_name:
            worker["type"] = "secure_patch"
        elif "silence" in patch_name:
            worker["type"] = "silence_patch"
        else:
            worker["type"] = "sys_patch"
        worker["install_patch_name"] = patch_name
        # add for mark
        worker["pkg_name"] = "system_patch_update"
        worker["patchID"] = patchID
        worker["callback"] = None
        self.push_work(worker)

    def get_patch_id(self):
        try:
            installed_patchID = subprocess.check_output(["grep", "install_patch", self.patch_version,]).decode().split('\n')[0].split('=')[1]
        except Exception as e:
            installed_patchID = 0
        try:
            installed_secure_patchID = subprocess.check_output(["grep", "install_secure_patch", self.patch_version,]).decode().split('\n')[0].split('=')[1]
        except Exception as e:
            installed_secure_patchID = 0
        try:
            installed_silence_patchID = subprocess.check_output(["grep", "install_silence_patch", self.patch_version,]).decode().split('\n')[0].split('=')[1]
        except Exception as e:
            installed_silence_patchID = 0
        return int(installed_patchID), int(installed_secure_patchID), int(installed_silence_patchID)

    #导入离线补丁
    def do_import_offline_patch(self, worker):
        env = worker["env"]
        logging.info( "do_import_offline_patch: env:" + str(env))
        msg = { 'status': 'inprogress', 'percent': 0, 'action':'import', 'error':'' }
        self.patchStatusChanged(msg)
        self._backup_env(env)
        cmd = "DEBIAN_FRONTEND=noninteractive apt-get install /tmp/offline-patch/*nfsupdate*.deb -y > /tmp/do_import_offline_patch.log 2>&1"
        if os.system(cmd) != 0:
            logging.error("do_import_offline_patch: excute cmd failed, cmd=" + cmd)
            cmd = "DEBIAN_FRONTEND=noninteractive apt-get install -f -y >>  /tmp/do_import_offline_patch.log 2>&1"
            os.system(cmd)
            msg = { 'status': 'failed', 'percent': 100, 'action':'import', 'error':'install failed' }
            self.patchStatusChanged(msg)
        else:
            logging.info("do_import_offline_patch: excute cmd success,cmd=" + cmd)
            msg = { 'status': 'success', 'percent': 100, 'action':'import', 'error':'' }
            self.patchStatusChanged(msg)
        shutil.rmtree("/tmp/offline-patch")
        self._restore_env()

    def write_patch_info_to_file(self, patch_name):
        try:
            patch_name = str(patch_name)
            last_sysupdate_patch_version = patch_name[-1]
            patch_type = "nfspatch"
            package_name = "nfsupdate"
            content_temp = patch_name.split('-')
            if (content_temp[1] == package_name):
                pass
            elif (content_temp[1] == "secure"):
                patch_type = "secure" + '-' + patch_type
            elif (content_temp[1] == "silence"):
                patch_type = "silence" + '-' + patch_type
            else:
                raise Exception("patch is error, please check it.")

            last_sysupdate_patch_desc_file = "/usr/share/cdos-update/" + patch_type + "/patch"  + last_sysupdate_patch_version + "/conf.json"
            logging.debug("patch_version: " + last_sysupdate_patch_version + ", " + "patch_type: " + patch_type + ", " + "patch_info_file: " + last_sysupdate_patch_desc_file)
            if os.path.exists(last_sysupdate_patch_desc_file):
                with open(last_sysupdate_patch_desc_file, 'r') as f:
                    j = json.load(f)
                last_sysupdate_patch_desc = j['desc']
            patch_name = patch_name + "++" + last_sysupdate_patch_desc + "\n";
            with open(self.patch_info, "a") as file:
                file.write(patch_name)
        except Exception as e:
            logging.error("write_patch_info_to_file failed" + str(e))
            return
        
    #安装补丁服务器的patch.
    def do_install_patches(self, worker):
        type = worker["type"]
        logging.info("do_install_patches:" + worker["install_patch_name"] + " type: " + type)
        msg = { 'status': 'inprogress', 'percent': 0, 'action':'install', 'type': type, 'error':'' }
        self.patchStatusChanged(msg)
        current_cwd = os.getcwd()
        self._backup_env(worker["env"])
        path = self.patch_path_dict[type]
        installed_sys_patchID, installed_secure_patchID, installed_silence_patchID= self.get_patch_id()
        if type == "sys_patch":
            installed_patchID = installed_sys_patchID
            key_in_patch_version_file = "install_patch"
        elif type == "secure_patch":
            installed_patchID = installed_secure_patchID
            key_in_patch_version_file = "install_secure_patch"
        else:
            installed_patchID = installed_silence_patchID
            key_in_patch_version_file = "install_silence_patch"
        begin_install_patchID = installed_patchID + 1
        logging.info("do_install_patches: will install patch from patch" + str(begin_install_patchID))
        if not os.path.exists(self.patch_version):
            os.system('echo  "install_patch=0\ninstall_secure_patch=0\ninstall_silence_patch=0" > ' + self.patch_version)

        while begin_install_patchID <= worker["patchID"]:
            try:
                os.chdir(path + "patch" + str(begin_install_patchID))
                #logging.info("current in " + os.getcwd())
                #logging.info("get env USER:" + str(os.getenv("USER")) + " HOME:" + str(os.getenv("HOME")) + " PATH:" + str(os.getenv("PATH")))
                with open("conf.json") as cfg:
                    patch_info = json.load(cfg)
                cmd = "/bin/bash " + "install > /tmp/install" + type + str(begin_install_patchID) + ".log 2>&1"
                logging.info("do_install_patches will excute cmd:" + cmd )
                if os.system(cmd) == 0:
                    logging.info("do_install_patches: install patch success, the cmd:" + cmd)

                    #update installed patchID in self.patch_version
                    if patch_info["upgrade_version"].lower() == "true":
                        cmd = "sed -i '/" + key_in_patch_version_file + "/d' " + self.patch_version + "; echo " + key_in_patch_version_file + "=0 >> " + self.patch_version
                    else:
                        cmd = "sed -i '/" + key_in_patch_version_file + "/d' " + self.patch_version + "; echo " + key_in_patch_version_file + "=" + str(begin_install_patchID) + " >> " + self.patch_version
                    logging.info("do_install_patches will excute cmd:" + cmd)
                    os.system(cmd)

                    #refresh progress
                    pr = int((begin_install_patchID - installed_patchID)*100/(worker["patchID"] - installed_patchID))
                    if pr == 100:
                        # 记录patch name
                        patch_name = worker["install_patch_name"]
                        logging.info("patch_name:" + patch_name)
                        self.write_patch_info_to_file(patch_name)
                        msg = { 'status': 'success', 'percent': pr, 'action':'install', 'type': type, 'error':'' }
                        #如果补丁安装成功，即清除本次备份标记
                        # self.backup_flag = False
                    else:
                        msg = { 'status': 'inprogress', 'percent': pr, 'action':'install', 'type': type, 'error':'' }
                    self.patchStatusChanged(msg)

                    #rmdir successful patch
                    os.chdir(path)
                    shutil.rmtree(path + "patch" + str(begin_install_patchID))

                    begin_install_patchID = begin_install_patchID + 1
                    #add record flag in nfs/info & os-release
                    for line in open("/etc/nfs/info","r"):
                        if "BUILDTIME=" in line:
                            buildtime=line.split("=")[1].replace('\n','')

                    for line in open("/etc/os-patchversion","r"):
                        if "install_patch" in line:
                            patchid=int(line.split("=")[1].replace('\n',''))
                            if patchid < 10:
                                nfs_record="N0"+str(patchid)
                            else:
                                nfs_record="N"+str(patchid)
                        elif "install_secure_patch" in line:
                            patchid=int(line.split("=")[1].replace('\n',''))
                            if patchid < 10: 
                                secure_record="S0"+str(patchid)
                            else:
                                secure_record="S"+str(patchid)
                            
                    record_cmd_info = "sed -i s/SYSTEM_RECORD=.*/SYSTEM_RECORD="+buildtime+nfs_record+secure_record+"/ /etc/nfs/info"
                    record_cmd_release = "sed -i s/SYSTEM_RECORD=.*/SYSTEM_RECORD="+buildtime+nfs_record+secure_record+"/ /etc/os-release"
                    os.system(record_cmd_info)
                    os.system(record_cmd_release)
                else:
                    logging.error("do_install_patches: install patch" + str(begin_install_patchID) + " failed")
                    msg = { 'status': 'failed', 'percent': 100, 'action':'install', 'type': type, 'error':'call install failed' }
                    self.patchStatusChanged(msg)
                    break
            except Exception as e:
                logging.error("do_install_patches:" + str(e))
                msg = { 'status': 'failed', 'percent': 100, 'action':'install', 'type': type, 'error':str(e) }
                self.patchStatusChanged(msg)
                break
        os.chdir(current_cwd)
        self._restore_env()


    @dbus.service.method(dbus_iface, in_signature='', out_signature='i')
    def getStatus(self):
        logging.debug("getStatus:" + self._status)
        return self._status

    @dbus.service.signal(dbus_interface=dbus_iface, signature='a{sv}',)
    def backup_recover_status_changed(self, msg):
        logging.debug("patchStatusChanged:" + str(msg))
        if msg["status"] == "success" or msg["status"] == "failed":
            self.continue_work()
            
    @dbus.service.signal(dbus_interface=dbus_iface, signature='a{sv}',)
    def patchStatusChanged(self, msg):
        logging.debug("patchStatusChanged:" + str(msg))
        if msg["status"] == "success" or msg["status"] == "failed":
            self.continue_work()

    @dbus.service.signal(dbus_interface=dbus_iface, signature='isa{sv}',)
    def upgradeStatusChanged(self, statusCode, status_text, paramList={}):
        logging.debug("upgradeStatusChanged:" + str(statusCode) + " " + status_text + " " + str(paramList))

    @dbus.service.signal(dbus_interface=dbus_iface, signature='issasasia{sv}',)
    def statusChanged(self, statusCode, status_text, patchID, correct_install_patches_list, patchList, progress, paramList={}):
        logging.debug("statusChanged:" + str(statusCode) + " " + status_text + " " + patchID + " " + str(correct_install_patches_list) + " " + str(patchList) + " " + str(progress) + " " + str(paramList))
        if statusCode == PatchStatus.ALLINSTALLED:
            self.continue_work()
        if statusCode == PatchStatus.ERROR_INSTALL:
            self.continue_work()
        pass

    @dbus.service.method( dbus_interface=dbus_iface, in_signature='as', out_signature='a{sa{sv}}')
    def get_app_info(self, pkgs) :
        return self.aptAction.get_app_info(pkgs)

    @dbus.service.method( dbus_interface=dbus_iface, in_signature='', out_signature='a{sa{sv}}')
    def get_all_app_info(self):
        return self.aptAction.get_all_app_info()

    @dbus.service.method( dbus_interface=dbus_iface, in_signature='', out_signature='as')
    def get_all_package_names(self):
        return self.aptAction.get_all_package_names()

    @dbus.service.method( dbus_interface=dbus_iface, in_signature='a{sv}as', out_signature='')
    def upgrade_pkgs(self, env, pkgs):
        GLib.idle_add(self.aptAction.upgrade, env, pkgs)

    @dbus.service.method( dbus_interface=dbus_iface, in_signature='a{sv}as', out_signature='')
    def install_pkgs(self, env, pkgs):
        GLib.idle_add(self.aptAction.install, env, pkgs)

    # system update in cdos update，系统更新中的系统组件和补丁更新，是deb包更新.
    @dbus.service.method( dbus_interface=dbus_iface, in_signature='a{sv}asasss', out_signature='')
    def commit_system_updates(self, env, pkgs_install, pkgs_upgrade, getpatch_info_site_path, getpatch_site_path):
        worker = {}
        worker["env"] = env
        worker["action"] = "commit"
        worker["type"] = "system_update"
        worker["install_pkg_list"] = pkgs_install
        worker["upgrade_pkg_list"] = pkgs_upgrade
        # patches
        worker["patch_info_site_path"] = getpatch_info_site_path
        worker["patch_path"] = getpatch_site_path
        # add for mark
        worker["pkg_name"] = "system_update"
        worker["pkg_version"] = ""
        worker["callback"] = None
        self.push_work(worker)

    # app update in cdos update，系统更新中的应用更新，也是deb包更新.
    @dbus.service.method( dbus_interface=dbus_iface, in_signature='a{sv}asas', out_signature='')
    def commit_app_pkgs(self, env, pkgs_install, pkgs_upgrade):
        worker = {}
        worker["env"] = env
        worker["action"] = "commit"
        worker["type"] = "app_update"
        worker["type2"] = "cdos_update"
        worker["pid"] = ""
        worker["process_name"] = ""
        worker["install_pkg_list"] = pkgs_install
        worker["upgrade_pkg_list"] = pkgs_upgrade
        # add for mark
        worker["pkg_name"] = pkgs_install[0]
        worker["pkg_version"] = ""
        worker["callback"] = None
        self.push_work(worker)

    def push_work(self, worker, sleep_time=0):
        # 如果队列中已存在
        if worker in self.work_queue:
            logging.info( "worker is in queue:" + str(worker))
            return True, ""
        # 如果任务早已存, 系统组件更新不会重复发送
        if (self.current_worker is not None) and (self.current_worker["pkg_name"] == worker["pkg_name"]):
            # 正在working
            if self._commit_flag:
                return False, "work is commiting"
            # current is cdos update
            if ("type2" in self.current_worker) and (self.current_worker["type2"] == "cdos_update"):
                return False, "work is running as force mode"
            # app update, same action
            if (self.current_worker["action"] == worker["action"]):
                return False, "work is running, try it later"
        if worker["pkg_name"] in self.work_queue_pkg_name:
            worker_id = self.work_queue_pkg_name.index(worker["pkg_name"])
            # if current app update is from cdos update client
            if ("type2" in self.work_queue[worker_id]) and (self.work_queue[worker_id]["type2"] == "cdos_update"):
                return False, "your package is waiting as force mode"
            self.work_queue[worker_id] = worker
            return True, ""

        self.work_queue.append(worker)
        self.work_queue_pkg_name.append(worker["pkg_name"])
        logging.info("push work:" + str(worker))
        logging.info("current work queue" + str(self.work_queue))
        GLib.idle_add(self.do_work, sleep_time)
        return True, ""

    #cdos-api中的应用更新api使用，用户点击安装更新的时候将应用的更新任务放到事件队列下一个立马执行.
    def push_work_to_head(self, worker, sleep_time=0):
        # 如果队列中已存在
        if worker in self.work_queue:
            logging.info( "worker is in queue:" + str(worker))
            return True, ""
        # 如果任务早已存, 系统组件更新不会重复发送
        if (self.current_worker is not None) and (self.current_worker["pkg_name"] == worker["pkg_name"]):
            # 正在working
            if self._commit_flag:
                return False, "work is commiting"
            # current is cdos update
            if ("type2" in self.current_worker) and (self.current_worker["type2"] == "cdos_update"):
                return False, "work is running as force mode"
            # app update, same action
            if (self.current_worker["action"] == worker["action"]):
                return False, "work is running, try it later"
        if worker["pkg_name"] in self.work_queue_pkg_name:
            worker_id = self.work_queue_pkg_name.index(worker["pkg_name"])
            # if current app update is from cdos update client
            if ("type2" in self.work_queue[worker_id]) and (self.work_queue[worker_id]["type2"] == "cdos_update"):
                return False, "your package is waiting as force mode"
            self.work_queue.pop(worker_id)
            self.work_queue_pkg_name.pop(worker_id)

        self.work_queue.insert( 0, worker)
        self.work_queue_pkg_name.insert( 0, worker["pkg_name"])
        logging.info("push work to head:" + str(worker))
        logging.info("current work queue" + str(self.work_queue))
        GLib.idle_add(self.do_work, sleep_time)
        return True, ""

    #cdos-api中的应用更新api使用
    def push_work_later(self, worker):
        time.sleep(3)
        self.push_work(worker)

    def do_work(self, sleep_time=0):
        if not sleep_time == 0:
            time.sleep(sleep_time)
        if self._work_flag:
            logging.info( "do_work: work flag is true,return")
            return
        else:
            self._work_flag = True
            self.do_work2()

    def do_work2(self):
        # if no work,reset flag
        if len(self.work_queue) == 0:
            self._work_flag = False
            self.working = False
            logging.info("no work, reset flag and wait for new work")
            return
        try:
            self.current_worker = worker = self.work_queue.pop(0)
            self.work_queue_pkg_name.pop(0)
            logging.info("do work2:" + str(worker))
            if worker["action"] == "commit":
                self.do_commit_pkgs(worker)
            elif worker["action"] == "install_patch":
                GLib.idle_add(self.do_install_patches, worker)
                pass
            elif worker["action"] == "download_patch":
                self.do_download_patch(worker)
            elif worker["action"] == "import_offline_patch":
                GLib.idle_add(self.do_import_offline_patch, worker)
            elif worker["action"] == "download":
                self.do_download_app_pkg(worker)
            elif worker["action"] == "backup":
                GLib.idle_add(self.do_backup)
            elif worker["action"] == "recover":
                GLib.idle_add(self.do_recover)
                pass
        except Exception as e:
            logging.error("do work2 Exception:" + str(e))

    def do_download_patch(self, worker):
        env, install_patch_name = worker["env"], worker["install_patch_name"]
        logging.info( "do_download_patch: env:" + str(env) + " install_patch_name:" + str(install_patch_name))
        self._commit_flag = True
        pkgs_install = [install_patch_name]
        pkgs_upgrade = []
        GLib.idle_add(self.aptAction.commit, env, pkgs_install, pkgs_upgrade, None)

    #work action 为commit，包含系统组件和应用更新.
    def do_commit_pkgs(self, worker):
        env, pkgs_install, pkgs_upgrade = worker["env"], worker["install_pkg_list"], worker["upgrade_pkg_list"]
        logging.info( "do_commit_pkgs: env:" + str(env) + " pkgs_install:" + str(pkgs_install) + " pkgs_upgrade:" + str(pkgs_upgrade))
        # if app update api，cdos-api中的应用更新api
        if ("type" in worker) and (worker["type"] == "app_update") \
            and ("type2" in worker) and (worker["type2"] == "app_update_api"):
            logging.info( "app update api do_commit_pkgs")
            # app update
            logging.info( "app update")
            # if process is running
            process_running_flag = False
            if ("pid" in worker) and ("process_name" in worker) \
                and ("exit_flag" in worker) and (worker["exit_flag"] is True):
                status, cmdline = self.get_pid_status_names(worker["pid"])
                if (status == True) and (cmdline == worker["process_name"]):
                    logging.info( "reset flag and push worker again")
                    self.current_worker = None
                    self._work_flag = False
                    process_running_flag = True
                    GLib.idle_add(self.push_work_later, worker)
            if process_running_flag is False:
                self._commit_flag = True
                GLib.idle_add(self.aptAction.commit, env, pkgs_install, pkgs_upgrade, worker)
        else:
            # app update or system update
            self._commit_flag = True
            GLib.idle_add(self.aptAction.commit, env, pkgs_install, pkgs_upgrade, None)

    #更新源的cache.
    @dbus.service.method( dbus_interface=dbus_iface, in_signature='a{sv}', out_signature='')
    def update_cache(self, env):
        GLib.idle_add(self.aptAction.update_cache, env)

    #下载新的软件中心数据库，但是软件中心服务器配置可能需要更改.
    @dbus.service.method( dbus_interface=dbus_iface, in_signature='', out_signature='b')
    def update_software_database(self):
        status = self.soft_server.download_new_database()
        if status == True:
            try:
                self.aptAction.update_app_info()
            except Exception as e:
                pass
        return status

    # used in aptAction
    def on_apt_status_changed(self, msg):
        if self.current_worker["action"] == "download_patch":
            if msg["status"] == "success" or msg["status"] == "failed":
                self._commit_flag = False
            msg["action"] = "download"
            msg["type"] = self.current_worker["type"]
            self.patchStatusChanged(msg)

        else:
            msg["type"] = self.current_worker["type"]
            msg["first_deb"] = self.current_worker["pkg_name"]
            msg["deb_list"] = self.current_worker["install_pkg_list"]
            self.AptStatusChanged(msg)

    def continue_work(self):
        self.current_worker = None
        self._work_flag = False
        self.do_work()

    #deb包更新是apt状态变化.
    @dbus.service.signal( dbus_interface=dbus_iface, signature='a{sv}')
    def AptStatusChanged(self, msg):
        try:
            logging.info("AptStatusChanged:" + str(msg))
            #deb包更新完成后，执行补丁更新.
            if msg["status"] == "success":
                self._commit_flag = False
                if self.current_worker["type"] == "app_update":
                    self.continue_work()
                elif self.current_worker["type"] == "system_update":
                    patchids = self.get_patchlist_to_install( self.current_worker["patch_info_site_path"])#获取需要安装patch.
                    self.do_install_patches( self.current_worker["patch_path"], patchids)
            if msg["status"] == "failed":
                # ToDo: do something if status is failed
                self._commit_flag = False
                self.continue_work()
                pass
        except Exception as e:
            logging.error("AptStatusChanged  Exception: " + str(e))
            pass
        pass

    def on_cache_status_changed(self, msg):
        self.cacheStatusChanged(msg)

    @dbus.service.signal( dbus_interface=dbus_iface, signature='a{sv}')
    def cacheStatusChanged(self, msg):
        pass

    # app update api,cdos-api中的应用更新api使用.
    @dbus.service.method( dbus_interface=dbus_iface, in_signature='s', out_signature='bas')
    def checkAppUpdate(self, pkg_name):
        try:
            if self.cache is None:
                self.cache = MyCache(None)
            else:
                self.cache.open(None)
                self.cache._initDepCache()
        except Exception as e:
            logging.error('checkAppUpdate exception %s:  %s' % (str(e), traceback.format_exc()))
            return False, []
        version_list = []
        if not pkg_name in self.cache:
            logging.error("checkAppUpdate error:" + str(pkg_name) + " is not in cache")
            return False, []
        cur_ver = None
        if self.cache[pkg_name].is_installed:
            cur_ver = self.cache[pkg_name].installed.version

        pkg_version_list = self.cache[pkg_name].versions
        for ver in pkg_version_list:
            if ver.is_installed is True:
                continue
            if not ver.version in version_list:
                if cur_ver is None:
                    continue
                elif (apt_pkg.version_compare(ver.version, cur_ver) < 0) :
                    continue
                else:
                    insert_flag = False
                    for i in range(0, len(version_list)):
                        if (apt_pkg.version_compare(ver.version, version_list[i]) < 0):
                            version_list.insert(i, ver.version)
                            insert_flag = True
                            break
                    if insert_flag == False:
                        if len(version_list) == 0:
                            version_list.append(ver.version)
                        elif (apt_pkg.version_compare(ver.version, version_list[-1]) > 0):
                            version_list.append(ver.version)

        if len(version_list) == 0:
            logging.info("checkAppUpdate: " + str(pkg_name) + " has no version can be update")
            return False, []
        logging.info("checkAppUpdate: " + str(pkg_name) + " verisons: " + str(version_list))
        return True, version_list

    #cdos-api中的应用更新api使用.
    @dbus.service.method( dbus_interface=dbus_iface, in_signature='a{sv}', out_signature='bs')
    def downloadAppUpdate(self, download_meta_data):
        logging.info("downloadAppUpdate")
        if len(self.work_queue) >= 20:
            return False, "too many works, try it later"
        if not ("package_name" in download_meta_data and "version" in download_meta_data \
            and "env_home" in download_meta_data and "env_path" in download_meta_data \
            and "env_user" in download_meta_data and "env_display" in download_meta_data):
            logging.error("downloadAppUpdate Error: paramaters error")
            return False, "paramaters error"
        # package msg
        pkg_name = download_meta_data["package_name"]
        pkg_version = download_meta_data["version"]
        # env
        env_home = download_meta_data["env_home"]
        env_path = download_meta_data["env_path"]
        env_user = download_meta_data["env_user"]
        env_display = download_meta_data["env_display"]
        worker = {}
        env = {}
        env["HOME"] = env_home
        env["PATH"] = env_path
        env["USER"] = env_user
        env["DISPLAY"] = env_display
        worker["env"] = env
        worker["action"] = "download"
        worker["type"] = "app_update"
        worker["pkg_name"] = pkg_name
        worker["pkg_version"] = pkg_version
        return self.push_work(worker)

    def do_download_app_pkg(self, worker):
        GLib.idle_add(self.aptAction.download_pkg, worker["env"], worker["pkg_name"], worker["pkg_version"], self.send_download_signal)

    def send_download_signal(self, pkg_name, pkg_version, status, process, err_reason=""):
        self.appDownloadProgressChanged(pkg_name, pkg_version, status, process, err_reason)
        logging.debug("send_download_signal :" + pkg_name + " " + pkg_version + " " + status\
            + " " + str(process) + " " + err_reason)
        if status == "end":
            self.continue_work()

    #cdos-api中的应用更新api使用.
    @dbus.service.signal( dbus_interface=dbus_iface, signature='sssis')
    def appDownloadProgressChanged(self, pkg_name, pkg_version, status, process, err_reason):
        pass

    # app update in cdos update #cdos-api中的应用更新api使用.
    @dbus.service.method( dbus_interface=dbus_iface, in_signature='a{sv}', out_signature='bs')
    def installAppUpdate(self, install_meta_data):
        logging.error( "installAppUpdate")
        if len(self.work_queue) >= 20:
            return False, "too many works, try it later"
        if not ("package_name" in install_meta_data and "version" in install_meta_data \
            and "env_home" in install_meta_data and "env_path" in install_meta_data \
            and "env_user" in install_meta_data and "env_display" in install_meta_data \
            and "pid" in install_meta_data and "process_name" in install_meta_data):
            logging.error("installAppUpdate Error: paramaters error")
            return False, "paramaters error"
        # package msg
        pkg_name = install_meta_data["package_name"]
        pkg_version = install_meta_data["version"]
        # env
        env_home = install_meta_data["env_home"]
        env_path = install_meta_data["env_path"]
        env_user = install_meta_data["env_user"]
        env_display = install_meta_data["env_display"]

        worker = {}
        env = {}
        env["HOME"] = env_home
        env["PATH"] = env_path
        env["USER"] = env_user
        env["DISPLAY"] = env_display
        worker["env"] = env
        worker["action"] = "commit"
        worker["type"] = "app_update"
        worker["type2"] = "app_update_api"
        worker["pid"] = install_meta_data["pid"]
        worker["process_name"] = install_meta_data["process_name"]
        if ("exit_flag" in install_meta_data) and (type(install_meta_data["exit_flag"]) is dbus.Boolean):
            worker["exit_flag"] = bool(install_meta_data["exit_flag"])
        else:
            worker["exit_flag"] = False
        # get pkg name and version
        if not pkg_version == "":
            pkg_name_with_verison = pkg_name + "=" + pkg_version
        else:
            pkg_name_with_verison = pkg_name

        worker["install_pkg_list"] = [pkg_name_with_verison]
        worker["upgrade_pkg_list"] = []
        # add for mark
        worker["pkg_name"] = pkg_name
        worker["pkg_version"] = pkg_version
        worker["callback"] = self.send_install_signal
        return self.push_work_to_head(worker)

    def send_install_signal(self, pkg_name, pkg_version, status, process, err_reason=""):
        self.appInstallProgressChanged(pkg_name, pkg_version, status, process, err_reason)
        logging.debug("send_install_signal :" + pkg_name + " " + pkg_version + " " + status\
            + " " + str(process) + " " + err_reason)

    #cdos-api中的应用更新api使用.
    @dbus.service.signal( dbus_interface=dbus_iface, signature='sssis')
    def appInstallProgressChanged(self, pkg_name, pkg_version, status, process, err_reason):
        pass


    def get_pid_status_names(self, pid):
        if str(pid) == "":
            return False, ""
        pid_proc_path = os.path.join("/proc", str(pid))
        pid_cmd_file_path = os.path.join(pid_proc_path, "cmdline")
        if os.path.isfile(pid_cmd_file_path):
            try:
                pid_cmd_fd = open(pid_cmd_file_path)
                pid_cmd = pid_cmd_fd.read()
                pid_cmd_fd.close()
                if pid_cmd.endswith("\x00"):
                    pid_cmd = pid_cmd[:-1]
                return True, pid_cmd
            except Exception as e:
                return False, ""
        else:
            return False, ""

    # app update in cdos update  #cdos-api中的应用更新api使用.
    @dbus.service.method( dbus_interface=dbus_iface, in_signature='s', out_signature='bs')
    def cancelAppUpdate(self, pkg_name):
        if (self.current_worker is not None) and (self.current_worker["pkg_name"] == pkg_name):
            return False, "work is running"
        while pkg_name in self.work_queue_pkg_name:
            pkg_id = self.work_queue_pkg_name.index(pkg_name)
            self.work_queue_pkg_name.pop(pkg_id)
            self.work_queue.pop(pkg_id)
        return True, ""

if __name__ == '__main__':
    # using glib
    import dbus.mainloop.glib
    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
    from gi.repository import GObject
    GObject.threads_init()
    loop = GObject.MainLoop()
    updatedaemon = CDOSUpdateDaemon()
    loop.run()

